from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Tuple

from mipa.ext.commands._types import _BaseCommand
from mipa.ext.commands.core import Command

if TYPE_CHECKING:
    from mipa.ext import Bot


class CogMeta(type):
    __cog_name__: str

    def __new__(cls, *args: Tuple[Any], **kwargs: Dict[str, Any]):
        name, bases, attrs = args
        attrs['__cog_name__'] = kwargs.pop('name', name)
        attrs['__cog_settings__'] = kwargs.pop('command_attrs', {})
        listeners = {}
        commands = {}
        no_bot_cog = 'Commands or listeners must not start with cog_ or bot_ (in method {0.__name__}.{1})'
        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)

        for base in reversed(new_cls.__mro__):  # 多重継承を確認 !コマンドを登録
            for elem, value in base.__dict__.items():
                if elem in commands:
                    del commands[elem]  # commandsから削除
                if elem in listeners:
                    del listeners[elem]  # listenersから削除
                is_static_method = isinstance(value, staticmethod)

                if isinstance(value, Command):
                    commands[elem] = value

                if is_static_method:  # staticmethodか確認
                    value = value.__func__  #
                # 関数をvalueに !valueが重要
                if isinstance(value, _BaseCommand):
                    commands[elem] = value
                elif inspect.iscoroutinefunction(value):
                    try:
                        value.__cog_listener__
                    except AttributeError:
                        continue
                    else:
                        if elem.startswith(('cog', 'bot_')):
                            raise TypeError(no_bot_cog.format(base, elem))
                        listeners[elem] = value
        new_cls.__cog_commands__ = list(commands.values())

        listeners_as_list: List[tuple[str, Any]] = []
        for listener in listeners.values():
            for listener_name in listener.__cog_listener_names__:
                listeners_as_list.append((listener_name, listener))

        new_cls.__cog_listeners__ = listeners_as_list
        return new_cls

    def __init__(self, *args: Tuple[Any], **kwargs: Dict[str, Any]):
        super().__init__(*args, **kwargs)

    @classmethod
    def qualified_name(cls) -> str:
        return cls.__cog_name__


class Cog(metaclass=CogMeta):
    __cog_name__ = ClassVar[str]
    __cog_settings__: Dict[str, Any] = {}
    __cog_listeners__: ClassVar[List[Tuple[str, str]]]
    __cog_commands__: List[Command] = []

    def __new__(cls, *args: tuple[Any], **kwargs: Dict[str, Any]):
        self = super().__new__(cls)
        self.__cog_commands__ = tuple(cls.__cog_commands__)
        return self

    @classmethod
    def listener(cls, name: Optional[str] = None):
        def decorator(func: Cog):
            actual = func
            if isinstance(actual, staticmethod):
                actual = actual.__func__
            if not inspect.iscoroutinefunction(actual):
                raise TypeError(
                    'Listener function must be a coroutine function.'
                )
            actual.__cog_listener__ = True
            to_assign = name or actual.__name__
            try:
                actual.__cog_listener_names__.append(to_assign)
            except AttributeError:
                actual.__cog_listener_names__ = [to_assign]
            return func

        return decorator

    @classmethod
    def qualified_name(cls) -> str:
        return cls.__cog_name__

    def _inject(self, bot: Bot):
        for command in self.__cog_commands__:
            bot.add_command(command, self.__cog_name__)

        for name, method_func in self.__cog_listeners__:
            bot.add_listener(getattr(self, name), name)

        return self



"""Misskeyのチャンネルへの接続や、メッセージのキャプチャ等のWebSocket関連"""

from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Iterable, Literal

if TYPE_CHECKING:
    from aiohttp.client_ws import ClientWebSocketResponse

__all__ = ['Router']


class Router:
    """
    Attributes
    ----------
    web_socket : ClientWebSocketResponse
        WebSocketクライアント

    Methods
    -------
    channels:
        与えられたlistを元にチャンネルに接続します
    global_time_line:
        WebSocketでGlobalTimeLineに接続します
    main_channel:
        WebSocketでMainチャンネルに接続します
    home_time_line:
        WebSocketでHomeTimeLineに接続します
    local_time_line:
        WebSocketでLocalTimeLineに接続します
    capture_message:
        与えられたメッセージを元にnote idを取得し、そのメッセージをon_message等の監視対象に追加します
    """

    def __init__(self, web_socket: ClientWebSocketResponse):
        self.web_socket: ClientWebSocketResponse = web_socket

    async def connect_channel(
        self,
        channel_list: Iterable[Literal['global', 'main', 'home', 'local']],
    ) -> None:
        """
        与えられたlistを元にチャンネルに接続します

        Parameters
        ----------
        channel_list : Iterable[Literal['global', 'main', 'home', 'local']]
            ['global', 'local', 'home', 'main']
        """

        channel_dict = {
            'global': 'globalTimeline',
            'main': 'main',
            'home': 'homeTimeline',
            'local': 'localTimeline',
        }
        try:
            for channel in channel_list:
                get_channel = channel_dict[channel]
                await self.web_socket.send_json(
                    {
                        'type': 'connect',
                        'body': {
                            'channel': f'{get_channel}',
                            'id': f'{uuid.uuid4()}',
                        },
                    }
                )

        except KeyError:
            pass

    async def capture_message(self, message_id: str) -> None:
        """
        与えられたメッセージを元にnote idを取得し、そのメッセージをon_message等の監視対象に追加します

        Parameters
        ----------
        message_id : str
        """

        await self.web_socket.send_json(
            {'type': 'subNote', 'body': {'id': f'{message_id}'}}
        )



__title__ = 'mipa'
__author__ = 'yupix'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022-present yupix'
__author_email__ = 'yupi0982@outlook.jp'
__version__ = '0.0.2'

__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from .client import Client
from .ext import *

__all__ = ('Client',)



"""MiPAのWebSocket部分"""

import json
import sys
from typing import Any, Optional

import aiohttp

from mipa import __version__

__all__ = ('HTTPClient', 'HTTPSession')


class _MissingSentinel:
    def __eq__(self, other):
        return False

    def __bool__(self):
        return False

    def __repr__(self):
        return '...'


MISSING: Any = _MissingSentinel()


async def json_or_text(response: aiohttp.ClientResponse):
    text = await response.text(encoding='utf-8')
    try:
        if 'application/json' in response.headers['Content-Type']:
            return json.loads(text)
    except KeyError:
        pass


class HTTPClient:
    def __init__(self) -> None:
        self.__session: aiohttp.ClientSession = MISSING
        self.token: Optional[str] = None
        user_agent = 'Misskey Bot (https://github.com/yupix/MiPA {0}) Python/{1[0]}.{1[1]} aiohttp/{2}'
        self.user_agent = user_agent.format(
            __version__, sys.version_info, aiohttp.__version__
        )

    async def close_session(self):
        await self.__session.close()

    async def ws_connect(self, url: str, *, compress: int = 0) -> Any:
        kwargs = {
            'autoclose': False,
            'max_msg_size': 0,
            'timeout': 30.0,
            'headers': {'User-Agent': self.user_agent},
            'compress': compress,
        }
        try:
            ws = await self.__session.ws_connect(url, **kwargs)
        except aiohttp.client_exceptions.ClientConnectorError:
            raise ClientConnectorError()
        return ws


HTTPSession: HTTPClient = HTTPClient()



__all__ = (
    'WebSocketRecconect',
    'CogNameDuplicate',
    'ExtensionAlreadyLoaded',
    'ExtensionFailed',
    'InvalidCogPath',
    'NoEntryPointError',
    'ClientConnectorError',
    'TaskNotRunningError',
)


class WebSocketRecconect(Exception):
    """Websocketに再接続が必要"""


class CogNameDuplicate(Exception):
    """Cogの名前が重複している"""


class ExtensionAlreadyLoaded(Exception):
    """Cogは既に読み込まれている"""


class ExtensionFailed(Exception):
    """Cog周りのエラー"""


class InvalidCogPath(Exception):
    """無効なCogのパス"""


class NoEntryPointError(Exception):
    """Cogにsetup関数が存在しない"""


class ClientConnectorError(Exception):
    """WebSocketの接続に問題が発生した"""


class TaskNotRunningError(Exception):
    """タスクが動いてない状態で停止しようとした"""



import pathlib

from setuptools import setup

description = 'A Python wrapper for the Misskey API'
readme_file = pathlib.Path(__file__).parent / 'README.md'
with readme_file.open(encoding='utf-8') as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as f:
    requirements = f.read().splitlines()

extras_require = {
    'dev': ['axblack', 'isort', 'mypy', 'flake8'],
    'ci': ['flake8', 'mypy'],
}

packages = ['mipa', 'mipa.ext', 'mipa.ext.commands', 'mipa.ext.tasks']

setup(
    name='mipa',
    version='0.0.2',
    install_requires=requirements,
    url='https://github.com/yupix/MiPA',
    author='yupix',
    author_email='yupi0982@outlook.jp',
    license='MIT',
    python_requires='>=3.10, <4.0',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=packages,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python :: 3.10',
        'Natural Language :: Japanese',
        'License :: OSI Approved :: MIT License',
    ],
    extras_require=extras_require,
)



from __future__ import annotations

from typing import TYPE_CHECKING, Generator, Optional

from mipac.core.models import RawUser
from mipac.http import Route
from mipac.models import User
from mipac.util import remove_dict_empty

if TYPE_CHECKING:
    from mipa.state import ConnectionState

__all__ = ('InstanceIterator',)


class InstanceIterator:
    def __init__(self, state: ConnectionState):
        self._state = state

    async def get_users(
        self,
        limit: int = 10,
        *,
        offset: int = 0,
        sort: Optional[str] = None,
        state: str = 'all',
        origin: str = 'local',
        username: Optional[str] = None,
        hostname: Optional[str] = None,
        get_all: bool = False
    ) -> Generator[User]:
        """
        Parameters
        ----------
        limit: int
        offset:int
        sort:str
        state:str
        origin:str
        username:str
        hostname:str
        get_all:bool

        Returns
        -------
        Iterator[User]
        """
        args = remove_dict_empty(
            {
                'limit': limit,
                'offset': offset,
                'sort': sort,
                'state': state,
                'origin': origin,
                'username': username,
                'hostname': hostname,
            }
        )
        res = await self._state.http.request(
            Route('POST', '/api/admin/show-users'),
            json=args,
            auth=True,
            lower=True,
        )

        if get_all:
            while True:
                for i in res:
                    yield User(RawUser(i))
                args['offset'] = args['offset'] + len(res)
                res = await self._state.http.request(
                    Route('POST', '/api/admin/show-users'),
                    json=args,
                    auth=True,
                    lower=True,
                )
                if len(res) == 0:
                    break
        else:
            for i in res:
                yield User(RawUser(i))



from __future__ import annotations

import asyncio
import inspect
from typing import TYPE_CHECKING, Any, Callable, Dict

from mipac.core.models import RawChat, RawNote, RawReaction, RawUser
from mipac.models import Chat, Emoji, FollowRequest, Note, Reaction, User
from mipac.types import ChatPayload, NotePayload, UserPayload
from mipac.util import str_lower, upper_to_lower

if TYPE_CHECKING:
    from mipa.client import Client


class ConnectionState:
    def __init__(
        self,
        dispatch: Callable[..., Any],
        loop: asyncio.AbstractEventLoop,
        client: Client,
    ):
        self.__client: Client = client
        self.__dispatch = dispatch
        # self.logger = get_module_logger(__name__) TODO: 直す
        self.loop: asyncio.AbstractEventLoop = loop
        self.parsers = parsers = {}
        for attr, func in inspect.getmembers(self):
            if attr.startswith('parse'):
                parsers[attr[6:].upper()] = func

    def parse_emoji_added(self, message: Dict[str, Any]):
        self.__dispatch('emoji_add', Emoji(message['body']['emoji']))

    def parse_channel(self, message: Dict[str, Any]) -> None:
        """parse_channel is a function to parse channel event

        チャンネルタイプのデータを解析後適切なパーサーに移動させます

        Parameters
        ----------
        message : Dict[str, Any]
            Received message
        """
        base_msg = upper_to_lower(message['body'])
        channel_type = str_lower(base_msg.get('type'))
        # self.logger.debug(f'ChannelType: {channel_type}') TODO: 修正
        # self.logger.debug(f'recv event type: {channel_type}') TODO: 修正
        getattr(self, f'parse_{channel_type}')(base_msg['body'])

    def parse_renote(self, message: Dict[str, Any]):
        pass

    def parse_unfollow(self, message: Dict[str, Any]):
        """
        フォローを解除した際のイベントを解析する関数
        """

    def parse_signin(self, message: Dict[str, Any]):
        """
        ログインが発生した際のイベント
        """

    def parse_receive_follow_request(self, message: Dict[str, Any]):
        """
        フォローリクエストを受け取った際のイベントを解析する関数
        """

        self.__dispatch('follow_request', FollowRequest(message))

    def parse_me_updated(self, message: UserPayload):
        self.__dispatch(
            'me_updated', User(RawUser(message), client=self.__client.action)
        )

    def parse_read_all_announcements(self, message: Dict[str, Any]) -> None:
        pass  # TODO: 実装

    def parse_reply(self, message: NotePayload) -> None:
        """
        リプライ
        """
        self.__dispatch(
            'message', Note(RawNote(message), client=self.__client.action)
        )

    def parse_follow(self, message: Dict[str, Any]) -> None:
        """
        ユーザーをフォローした際のイベントを解析する関数
        """

        self.__dispatch(
            'user_follow', User(RawUser(message), client=self.__client.action)
        )

    def parse_followed(self, message: Dict[str, Any]) -> None:
        """
        フォローイベントを解析する関数
        """

        self.__dispatch(
            'follow', User(RawUser(message), client=self.__client.action)
        )

    def parse_mention(self, message: Dict[str, Any]) -> None:
        """
        メンションイベントを解析する関数
        """

        self.__dispatch(
            'mention', Note(RawNote(message), client=self.__client.action)
        )

    def parse_drive_file_created(self, message: Dict[str, Any]) -> None:
        self.__dispatch('drive_file_created', message)

    def parse_read_all_unread_mentions(self, message: Dict[str, Any]) -> None:
        pass  # TODO:実装

    def parse_read_all_unread_specified_notes(
        self, message: Dict[str, Any]
    ) -> None:
        pass  # TODO:実装

    def parse_read_all_channels(self, message: Dict[str, Any]) -> None:
        pass  # TODO:実装

    def parse_read_all_notifications(self, message: Dict[str, Any]) -> None:
        pass  # TODO:実装

    def parse_url_upload_finished(self, message: Dict[str, Any]) -> None:
        pass  # TODO:実装

    def parse_unread_mention(self, message: Dict[str, Any]) -> None:
        pass

    def parse_unread_specified_note(self, message: Dict[str, Any]) -> None:
        pass

    def parse_read_all_messaging_messages(
        self, message: Dict[str, Any]
    ) -> None:
        pass

    def parse_messaging_message(self, message: ChatPayload) -> None:
        """
        チャットが来た際のデータを処理する関数
        """
        self.__dispatch(
            'message', Chat(RawChat(message), client=self.__client.action)
        )

    def parse_unread_messaging_message(self, message: Dict[str, Any]) -> None:
        """
        チャットが既読になっていない場合のデータを処理する関数
        """
        self.__dispatch(
            'message', Chat(RawChat(message), client=self.__client.action)
        )

    def parse_notification(self, message: Dict[str, Any]) -> None:
        """
        通知イベントを解析する関数

        Parameters
        ----------
        message: Dict[str, Any]
            Received message
        """

        accept_type = ['reaction']
        notification_type = str_lower(message['type'])
        if notification_type in accept_type:
            getattr(self, f'parse_{notification_type}')(message)

    def parse_follow_request_accepted(self, message: Dict[str, Any]) -> None:
        pass

    def parse_poll_vote(self, message: Dict[str, Any]) -> None:
        pass  # TODO: 実装

    def parse_unread_notification(self, message: Dict[str, Any]) -> None:
        """
        未読の通知を解析する関数

        Parameters
        ----------
        message : Dict[str, Any]
            Received message
        """
        # notification_type = str_lower(message['type'])
        # getattr(self, f'parse_{notification_type}')(message)

    def parse_reaction(self, message: Dict[str, Any]) -> None:
        """
        リアクションに関する情報を解析する関数
        """
        self.__dispatch(
            'reaction',
            Reaction(RawReaction(message), client=self.__client.action),
        )

    def parse_note(self, message: NotePayload) -> None:
        """
        ノートイベントを解析する関数
        """
        note = Note(RawNote(message), self.__client.action)
        # Router(self.http.ws).capture_message(note.id) TODO: capture message
        self.__client._on_message(note)



from __future__ import annotations

import asyncio
import functools
from typing import TYPE_CHECKING, List, Optional

from mipa.ext.commands._types import _BaseCommand

if TYPE_CHECKING:
    from mipa.ext.commands import Context


def hooked_wrapped_callback(coro):
    @functools.wraps(coro)
    async def wrapped(*args, **kwargs):
        ret = await coro(*args, **kwargs)
        return ret

    return wrapped


class CMD:
    def __init__(
        self, cmd_type: str, key: str, func: 'Command', cog_name: str
    ):
        self.cmd_type = cmd_type
        self.key = key
        self.func = func
        self.cog_name = cog_name


class CommandManager:
    def __init__(self, *args, **kwargs):
        self.all_commands: List[CMD] = []
        super().__init__(*args, **kwargs)  # Clientクラスを初期化する

    def add_command(self, command: 'Command', cog_name: str):
        if not isinstance(command, Command):
            raise TypeError(f'{command}はCommandクラスである必要があります')
        command_type = 'regex' if command.regex else 'text'
        command_key = command.regex or command.text
        self.all_commands.append(
            CMD(command_type, command_key, command, cog_name)
        )


class Command(_BaseCommand):
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        return self

    def __init__(self, func, regex: str, text: str, **kwargs):
        if not asyncio.iscoroutinefunction(func):
            raise TypeError(f'{func}はコルーチンでなければなりません')
        self.regex: str = regex
        self.text: str = text
        self.callback = func
        self.cog = None

    @property
    def qualified_name(self) -> str:
        return self.regex or self.text

    def __str__(self):
        return self.qualified_name

    @staticmethod
    async def _parse_arguments(ctx: Context):
        args = [ctx] if ctx.cog is None else (ctx.cog, ctx)
        ctx.args = args + ctx.args
        return ctx

    async def invoke(self, ctx: Context, *args, **kwargs):
        ctx = await self._parse_arguments(ctx)
        await self.callback(*ctx.args, **ctx.kwargs)


def mention_command(regex: Optional[str] = None, text: Optional[str] = None):
    def decorator(func, **kwargs):
        return Command(func, regex=regex, text=text, **kwargs)

    return decorator
